from .Cluster import FileCluster
from .Association import AssociationCluster
from .Cache import AutoCache
from .Broadcast import Broadcaster
from .cogs.VariableParser import VariableParser 
from .cogs.VariableAction import VariableAction



class FileClusterIterator:
    '''
        A simple iterator to run through the variable data
    '''
    def __init__(self, variable):
        self.var_vals = variable.vals()
        self.index = -1

    def __next__(self):
        """
        This is an implementation of the `__next__` method for an iterator that raises a `StopIteration`
        exception when the end of the iterator is reached.
        :return: The `__next__` method is returning the next value in the `var_vals` list after
        incrementing the `index` by 1.
        """

        if self.index + 1 >= len(self.var_vals):
            raise StopIteration
        
        self.index += 1
        r = self.var_vals[self.index]
        return r


from pathlib import Path
from functools import wraps
from refyre.utils import get_optimal_pattern, optional_dependencies
from refyre.config import logger
from refyre.reader import PatternGenerator
import re

#Copying / Deleting
import shutil

#Iteration
from .ClusterIterator import FileClusterIterator
from .Broadcast import Broadcaster
from .Cache import AutoCache

#Copy
import copy

#Zip
import zipfile


#Store a weak reference to the object
import weakref


def handle_file_conflict(destination_path):
    if not destination_path.exists():
        return destination_path

    base_name = destination_path.stem
    suffix = destination_path.suffix
    directory = destination_path.parent

    index = 1
    while True:
        new_name = f"{base_name}({index}){suffix}"
        new_path = directory / new_name
        if not new_path.exists():
            return new_path
        index += 1

class FileCluster:
    '''
        The workhorse of refyre. Variables do all of the heavylifting, harnessing
        the power of the data parsed through the fgraphs.

        Paradigms - any method that involves transporting data must be wrapped in the 
        refresher decorator method above. This will prevent overlapping variables from 
        incorrectly altering results from each other, as well as ensures that variables
        only contain valid data.
    '''

    #Counter to manage IDs
    GLOBAL_COUNTER = 0

    #Stores references to each FileCluster; in the future, this can be used for tracking & mem management
    clusters = {}

    def __init__(self, input_paths = [], input_patterns = [], values = [], as_pathlib = False, recursive = False):
        assert len(input_paths) == len(input_patterns), "Uneven lengths for input paths and patterns"
        logger.debug('init')

        for pattern in input_patterns:
            logger.debug(f'add {pattern} {PatternGenerator(pattern)}')
        self.values = self.__populate(input_paths, [ PatternGenerator(p) for p in input_patterns  ], recursive = recursive)

        #ID the variable
        self.id = FileCluster.GLOBAL_COUNTER
        FileCluster.GLOBAL_COUNTER += 1

        #Track the cluster without affecting garbage collection
        self.weak_ref = weakref.ref(self)
        FileCluster.clusters[self.id] = self.weak_ref

        #Setup a cache dir
        self.cache_dir = '.refyre_cache'
    
        #If the user didn't send the values as Pathlib objects already
        if not as_pathlib: 
            self.values += [ Path(pth) for pth in values ]
        else:
            self.values += values

        self.values = sorted(self.values)
        #Track variable on broadcaster
        Broadcaster.add(self.id, self.values)

    def all_clusters(self):
        return FileCluster.clusters

    def __del__(self):
        if self.id in FileCluster.clusters:
            FileCluster.clusters.pop(self.id)
        Broadcaster.release(self.id)

    
    @classmethod
    def cleanup(cls):
        for k in list(FileCluster.clusters):
            obj = FileCluster.clusters[k]
            if obj() is None:
                FileCluster.clusters.pop(k)

    @classmethod
    def wipe(cls):
        for k in list(FileCluster.clusters):
            logger.debug(f'wiping {k}')
            obj = FileCluster.clusters[k]()
            if obj is not None:
                logger.debug(f'deleting {k}')
                del obj 
        
        FileCluster.GLOBAL_COUNTER = 0
        Broadcaster.clear()
        FileCluster.clusters.clear()

    def __add__(self, other):
        '''
            other - another FileCluster

            Returns a file cluster with all the combined values
        '''
        #Ensure the input_data is as desired
        assert type(other) == FileCluster

        new_values = self.values + other.values

        return FileCluster(input_paths = [], input_patterns = [], values = new_values, as_pathlib = True)
    
    def __and__(self, other):
        '''
            other - another FileCluster

            Returns the intersection between the two FileClusters
        '''

        return FileCluster(input_paths = [], input_patterns = [], values = list(set(self.values).intersection(other.values)))
    
    def __or__(self, other):
        '''
            other - another FileCluster

            Returns the union between two FileClusters
        '''

        return FileCluster(input_paths = [], input_patterns = [], values = list(set(self.values).union(other.values)))
    
    def __xor__(self, other):
        '''
            other - another FileCluster

            Returns the exclusive or between the two FileClusters
        '''
        return FileCluster(input_paths = [], input_patterns = [], values = list(set(self.values) ^ set(other.values)))
    
    def __sub__(self, other):
        '''
            other - another FileCluster

            Returns all the files that are exclusively in self
        '''

        return FileCluster(input_paths = [], input_patterns = [], values = [v for v in self.values if v not in other.values])

    def __eq__(self, other):
        if other is None or type(other) != FileCluster:
            return False
        return set(self.values) == set(other.values)
    
    def __contains__(self, other):
        return self & other == other

    def __populate(self, input_paths, input_patterns, recursive = False):
        '''
        Populates a variable, using the input paths and directories 
        '''

        ret = []
        for dir_path, pattern in zip(input_paths, input_patterns):
            logger.debug(f'pot {dir_path} {pattern}')

            it = sorted(Path(dir_path).iterdir()) if not recursive else sorted(Path(dir_path).glob('**/*'))
            for fl in it:
                if re.search(r'{}'.format(pattern), fl.name) and fl not in ret:
                    ret.append(fl)
        
        return ret

    def vals(self):
        '''
            Returns a Pathlib.Paths list of all
            filepaths
        '''
        return self.values
    
    def item(self):
        '''
            Returns a Pathlib.Paths list of all
            filepaths
        '''
        return self.values[0]

    def dirs(self):
        '''
            Returns a FileCluster of all parent directories
        '''
        return FileCluster(values = list(dict.fromkeys([v.parent for v in self.values])), as_pathlib = True)
    
    def __repr__(self):
        return f"FileCluster(values = {self.values})"

    def __len__(self):
        return len(self.values)

    def __iter__(self):
        return FileClusterIterator(self)

    def __copy__(self):
        return FileCluster(values = self.values, as_pathlib = True)
    
    def __deepcopy__(self):
        return FileCluster(values = [Path(p) for p in self.values], as_pathlib = True)

    def __getitem__(self, key):
        if isinstance(key, slice):
            return FileCluster(input_patterns = [] , input_paths = [], values = self.values[key.start:key.stop:key.step], as_pathlib = True)
        elif isinstance(key, int):
            return FileCluster(input_patterns = [] , input_paths = [], values = [self.values[key]], as_pathlib = True)

    @Broadcaster()
    def move(self, target_dir, conflict_function = handle_file_conflict):
        """
        The above code defines functions for moving, copying, and deleting files in a directory, with
        options for handling conflicts and broadcasting changes.
        
        :param target_dir: The directory to which files will be moved or copied
        :param conflict_function: conflict_function is a function that takes in a Path object
            representing a file and returns a Path object representing the destination file path. This
            function is used in the move and copy methods to handle file conflicts, such as when a file with
            the same name already exists in the target directory. The default conflict
        :return: The `move` method returns a tuple containing a list of changes made (as tuples of old
        and new file paths) and a new `FileCluster` object with updated values.
        """


        p = Path(target_dir)

        #If the directory doesn't actually exist or just isn't a directory, we return None
        if not p.exists() or not p.is_dir():
            return None

        nvals = []
        changes = []

        for v in self.values:
            logger.debug(f'{v.parent}, {p}')
            if v.parent != p:
                logger.debug('in')
                dest = conflict_function(p / v.name)

                #if dest is None, we ignore it
                if dest:
                    logger.debug(f'{str(v)}, {str(dest)}')
                    shutil.move(str(v), str(dest))
                    changes.append((str(v), str(dest)))
                    nvals.append(Path(dest))
            else:
                logger.debug('else')
                logger.debug(str(v))
                changes.append((str(v), str(v)))
                nvals.append(Path(v))
            
        return changes, FileCluster(input_patterns = [], input_paths = [], values = nvals, as_pathlib = True)
    
    def copy(self, target_dir, conflict_function = handle_file_conflict):
        '''
        Input: 
            - target_dir: the directory to which all files will be sent.
        
        Returns
            - FileCluster object with the new values 
        '''
        p = Path(target_dir)

        #If the directory doesn't actually exist or just isn't a directory, we return None
        if not p.exists() or not p.is_dir():
            return None
        
        nvals = []
        for v in self.values:
            dest = conflict_function(p / v.name)

            #if dest is None, we ignore it
            if dest:
                shutil.copy(str(v), str(dest))
                nvals.append(Path(dest))
            
        return FileCluster(input_patterns = [], input_paths = [], values = nvals, as_pathlib = True)

    @Broadcaster()
    def delete(self):
        """
        The function deletes all files in a FileCluster object and returns an empty object while keeping
        track of the changes made.
        :return: a tuple containing two elements: a list of changes made during the deletion process and
        an empty FileCluster object.
        """
        changes = []
        for fl in self.values:
            if fl.is_dir():
                shutil.rmtree(str(fl))
                changes.append((str(fl), None))
            elif fl.is_file():
                fl.unlink()
                changes.append((str(fl), None))


        return changes, FileCluster(input_patterns = [], input_paths = [], values = [])

    @Broadcaster()
    def rename(self, rename_function):
        """
        This function renames files in a FileCluster object using a provided renaming function.
        
        :param rename_function: The `rename_function` parameter is a function that takes an integer as
            an input and returns a string. This function is used to rename the files in the `values` list of
            the `FileCluster` object. The integer input to the function is the index of the file in the
            `values`
        :return: a tuple containing two values: 
        1. A list of tuples representing the changes made during the renaming process. Each tuple
        contains two strings: the original path and the new path.
        2. A new instance of the FileCluster class with the updated values after the renaming process.
        """
        
        nvals = []
        changes = []
        for i, v in enumerate(self.values):
            logger.debug(f'renaming, {v},{v.parent / rename_function(i)}')
            nvals.append(Path(v).rename(v.parent / rename_function(i)))
            changes.append((str(v), str(nvals[-1])))

        return changes, FileCluster(input_patterns = [], input_paths = [], values = nvals, as_pathlib = True)

    def zip(self, save_dir = '.', save_name = "out.zip", conflict_function = handle_file_conflict):
        """
        This function zips all the files in a directory into a single zip file.
        
        :param save_dir: The directory where the zip file will be saved. If not specified, it will
        default to the current working directory, defaults to . (optional)
        :param save_name: The name of the zip file that will be created, defaults to out.zip (optional)
        :param conflict_function: The conflict_function parameter is a function that is used to handle
        conflicts that may arise when trying to save the zip file to the specified directory. It takes
        in the path of the file that is being saved and returns a new path if there is a conflict with
        an existing file in the directory. If no
        :return: a `FileCluster` object that contains the path to the newly created zip file.
        """

        save_p = Path(save_dir) / save_name

        copied = self.copy(save_dir)


        save_p = handle_file_conflict(save_p)
        with zipfile.ZipFile(str(save_p), 'w') as zipMe:        
            for fl in copied.values:
                zipMe.write(str(fl), compress_type=zipfile.ZIP_DEFLATED)
        
        assert save_p.exists(), "For some reason, the newly created zip file does not exist"

        copied.delete()

        return FileCluster(input_patterns = [], input_paths = [], values = [save_p], as_pathlib = True)

    def post(self, url, additional_data, payload_name):
        """
        This function sends a post request to a specified API endpoint with additional payload parameters
        and returns the response object.
        
        :param url: The link to the API endpoint where the post request will be sent
        :param additional_data: a dictionary containing any additional metadata you want to send along with
        the request. This could include things like authentication tokens or other parameters required by
        the API endpoint you are sending the request to
        :param payload_name: The name you want to give to the zip file that will be sent as part of the POST
        request
        :return: the Response object returned from the requests.post() call.
        """

        with optional_dependencies('warn'):
            #GET/POST requests
            import requests
            '''
            Uses the requests library to send a post request to the url, 
            alongside the additional payload parameters you specify. All the data 
            you want to push through is sent as a zip 

                url: Str link to the API endpoint
                additional_data: a dict to any other metadata you want to ferry across
                payload_name: what you want to name the zip going across


            Return type: the Response object returned from the requests.get() call
            '''

            #Zip all the files up to send
            zipped = self.zip()
            zipped_fp = str(zipped.item())

            resp = None
            with open(zipped_fp, "rb") as f:
                resp = requests.post(url = url, params = additional_data, files = { payload_name : f })

            #Delete the intermediate zip we constructed
            zipped.delete()

            return resp 

    #DO NOT ADD AN AUTOREFRESHER DECORATOR HERE --> It will add a middleman to move() and copy() that will alter results
    def map(self, map_func):
        """
        This function maps each filepath in a FileCluster object using a given function and returns a
        new FileCluster object with the mapped values.
        
        :param map_func: `map_func` is a function that takes a filepath as input and returns a filepath
        as output. This function is used to map each filepath in the `values` attribute of a
        `FileCluster` object. The resulting `FileCluster` object will have the same structure as the
        original one, but
        :return: A new FileCluster object with the mapped values.
        """
        nvals = []
        for v in self.values:
            nvals.append(Path(map_func(v))) #Append a copy of the object to prevent object ref shenanigans
        
        p = FileCluster(input_patterns = [], input_paths = [], values = nvals, as_pathlib = True)
        return p
    
     
    def reduce(self, reducer_function):
        """
        The reduce function takes in a reducer function and returns the result of reducing the values in
        the array using the reducer function, or None if the array is empty.
        
        :param reducer_function: A function that takes in two parameters and returns a third parameter
        that can be used to reduce a list of values into a single value. This function is used in the
        reduce method of a class to perform the reduction operation on the list of values stored in the
        class instance
        :return: the result of the reduction of the values in the array using the given reducer
        function. If the array is empty, it returns None.
        """
        if len(self.values):
            a = self.values[0]

            for i in range(1, len(self.values)):
                a = reducer_function(a, self.values[i])

            return a
        
        return None
    
    def filter(self, filter_func):
        """
        This function takes a filter function as input and returns a new FileCluster object with values
        filtered based on the input function.
        
        :param filter_func: filter_func is a function that takes a filepath as input and returns a boolean
        value. It is used to filter out values that we don't want to keep in a FileCluster object. If the
        function returns True for a given filepath, it will be included in the filtered FileCluster object.
        If it
        :return: A FileCluster object with the filtered values.
        """

        nvals = []
        for v in self.values:
            if filter_func(v):
                nvals.append(Path(v)) #Append a copy of the object to prevent object ref shenanigans
        
        return FileCluster(input_patterns = [], input_paths = [], values = nvals, as_pathlib = True)

    def clone(self):
        """
        The function returns a deep clone of the current object.
        :return: A deep clone of the current object is being returned.
        """

        return self.__deepcopy__()

    def decompress(self, complete = False):
        """
        Takes any folders within the values, and replaces them with all the items in the folder

        If complete is set to True, then it will recursively empty all files 
        """
        nvals = []
        for pth in self.values:
            if pth.is_dir():
                nvals.extend([*pth.iterdir()] if not complete else [*pth.rglob('**/*')])
            elif pth.is_file():
                nvals.append(pth)
        
        return FileCluster(values = nvals, as_pathlib = True)
            
    def cache(self):
        """
        Saves all the files to cache
        """

        ca = AutoCache(self.cache_dir)
        return ca.cache(self)
    
    def decache(self):
        ca = AutoCache(self.cache_dir)
        return ca.decache(self)
    
    def cached(self):
        """
        Returns all the FileCluster files in the cluster that are currently cached
        """
        nvals = []
        ca = AutoCache(self.cache_dir)

        for v in self.values:
            if ca.contains(str(v)):
                nvals.append(v)
            
        return FileCluster(values = nvals, as_pathlib = True)
    
    def unlink(self):
        """
        Returns a FileCluster detached from the Broadcaster system
        """

        f = self.clone()
        Broadcaster.unlink(f.id)
        return f
    
    def relink(self):
        """
        Reconnect the FileCluster to the broadcaster
        """
        return self.clone()

    def filesize(self, format = "bytes"):
        """
        Returns the total file size, in bytes 
        """
        total_size = 0.0

        for pth in self.values:
            if pth.is_dir():
                total_size += sum([p.stat().st_size for p in Path(pth).rglob('*')])
            elif pth.is_file():
                total_size += pth.stat().st_size

        return total_size

    def filecount(self):
        """
        Returns the total number of files over all folders
        """
        total_files = 0

        for pth in self.values:
            if pth.is_dir():
                total_files += len([p.stat().st_size for p in Path(pth).rglob('*')])
            elif pth.is_file():
                total_files += 1
        
        return total_files


from refyre.graph import FileGraph
from refyre.reader import Lexer, Parser, ExpressionGenerator, PatternGenerator 
from refyre.cluster import FileCluster, VariableParser, VariableAction
from refyre.utils import is_valid_regex, clone_node, extract_numbers
from refyre.config import logger
from refyre.core import CodeManager, AliasManager

#pathlib
from pathlib import Path

#Garbage collection
import gc

#Regex
import re

#Deleting
import shutil

#Globbing
import fnmatch

#json 
import json

class Refyre:

    def __init__(self, input_specs = None, output_specs = None, variables = None):
        '''
            input_specs: A List[Str] of filepaths to the various specs refyre will target
        '''

        self.variables = {} if not variables else variables

        #Initialize the fgraphs, proto and real

        #The proto 
        self.file_graph_blueprint = FileGraph() 
        self.file_graph = FileGraph()

        self.code_manager = CodeManager()
        self.alias_manager = AliasManager()

        if input_specs:
            assert isinstance(input_specs, list)
            for spec in input_specs:
                self.add_spec(spec)
        
        if output_specs:
            assert isinstance(output_specs, list)
            for spec in output_specs:
                self.create_spec(spec)

    def __getitem__(self, key):
        '''
            The refyre variables ("FileClusters") can be
            internally referred to through the Refyre object
        '''
        return self.variables[key] if key in self.variables else None
    
    def __setitem__(self, key, value):
        '''
            Attaches an externally made variable to the refyre object
        '''
        assert type(key) == str, "Key should be a string referring to a variable name"
        assert type(value) == FileCluster, "Value should be a FileCluster variable"
        self.variables[key] = value

    def __len__(self):
        return len(self.variables)

    def __contains__(self, key):
        return key in self.variables

    #The Five Fundamental Operations of Refyre
    def __construct(self, input_path, is_output = False, expand_path = ""):
        '''
            Construct Operation:
                - Receives an input spec, and creates an fgraph
        '''


        p = Parser(Lexer(input_path))

        logger.debug(f"Constructed graph: {p}")

        #Monkey patch, until we figure out how we can support directory clustering with output
        return p.copy(), self.__expand(p, path = expand_path)[0] 
        
    def __reconstruct(self, input_fgraph, expand_path = "", step = True):
        '''
            Reconstruct 
                - A mirror operation to the construct, except done with an
                fgraph as an input
        '''
        return input_fgraph.copy(), self.__expand(input_fgraph, path = expand_path, step = step)[0]


    def __expand(self, node, path = "", step = False):
        '''
            Expands all regex directories THAT EXIST into a static fgraph.

            If no such directories exist, we leave the node unaltered.

            This method enables a single cluster to target multiple directories
        '''
        if not node:
            return [None]

        new_path = Path(node.directory) if node.is_root_dir() else Path(path) / node.directory

        #assert Path(path).exists(), f"Error, the path {path} doesn't exist"
        logger.debug(f'new path: {new_path} {node.directory} {node.pattern}')
        
        if (Path(path).is_file()) or (not new_path.exists() and not PatternGenerator.is_valid_regex(node.directory)):
            logger.debug(f'invalid in general {Path(path)} {new_path} {not new_path.exists() and not PatternGenerator.is_valid_regex(node.directory)}')
            return [None]
        

        #Handle limits 
        has_limit, lower, upper = False, 0, 1000000000
        if node.limit != '':
            has_limit = True
            lower, upper = extract_numbers(node.limit)
        logger.debug(f"LIMIT STUFF {lower} {upper} {has_limit}")


        logger.debug(f"DIR {node.directory}")
        t = PatternGenerator.get_pattern_type(node.directory)
        logger.debug(f'detected pattern to be {t} {node.directory == ""}')

        if t != "normal_string":
            file_num = 0
            ret = []

            logger.info(t)

            if (t == "regex" or t == "glob") and node.directory != '':
                logger.info('valid regex')

                #Create nodes for each of the pattern matches
                logger.info(f'{new_path}, {new_path.parent}')
                for fl in new_path.parent.iterdir():

                    logger.info(f'file {fl}')
                    if re.search(r'{}'.format(PatternGenerator(node.directory)), str(fl)) and fl not in ret and fl.is_dir() and file_num >= lower and file_num <= upper: #No files will be allowed to be fclusters

                        pattern_matched_node = node.copy()
                        assert pattern_matched_node.pattern == node.pattern, f"Copy node has pattern {pattern_matched_node.pattern} while node has pattern {node.pattern}"

                        assert type(fl.name) == str
                        #Update the directory with the name 
                        pattern_matched_node.directory = fl.name 

                        logger.info(f'found match {fl.name}')
                        file_num += 1
                        ret.append(pattern_matched_node)                 
            
            elif t == "generator_expression":

                #If the limit is specified, we create a node for every number within the range
                gen_func = ExpressionGenerator(node.directory)

                mx_num = -1

                logger.info('\n\n\nGENERATOR!!!!', )

                if has_limit:
                    for j in range(lower, upper + 1):
                        new_node = node.copy()
                        new_node.directory = gen_func(j)
                        ret.append(new_node)
                    
                    mx_num = upper

                #If the limit isn't specified, we detect every node that follows the pattern                    
                else:
                    
                    for fl in new_path.parent.iterdir():

                        reversed_num = ExpressionGenerator.reverse_generator_expression(node.directory, fl.name)
                        logger.debug(f"DIR {node.directory} {str(fl)} {reversed_num}")
                        if reversed_num != None:
                            logger.debug("REVERSED")
                            mx_num = max(mx_num, reversed_num)

                            pattern_matched_node = node.copy()
                            assert pattern_matched_node.pattern == node.pattern, f"Copy node has pattern {pattern_matched_node.pattern} while node has pattern {node.pattern}"

                            assert type(fl.name) == str
                            #Update the directory with the name 
                            pattern_matched_node.directory = fl.name 

                            file_num += 1
                            ret.append(pattern_matched_node)  

                if step and '*s' in node.flags:
                    new_node = node.copy()

                    new_node.directory = gen_func(mx_num + 1)


                    #Take a pause to generate the new directory 
                    self.__output(new_node, new_path.parent, mode = "copy")


                    ret.append(new_node)

            
            #Before we go to children, let's handle any imports we must
            import_fgraph = None
            if node.imports != '' and Path(node.imports).exists():
                _, import_fgraph = self.__construct(node.imports, expand_path = new_path)
                import_fgraph.is_root = False

            #Now, attempt to check for all the node children of the original node
            #We will insert each of the current node's children into the new matched nodes, and recurse on them
            logger.debug(f'ret {ret}')

            for i, pattern_node in enumerate(ret):
                nchilds = []

                new_path = Path(pattern_node.directory) if pattern_node.is_root_dir() else Path(path) / pattern_node.directory 
                for child in node.children + [import_fgraph]:
                    nchilds.extend(self.__expand(child, new_path, step = step))

                pattern_node.children = [c for c in nchilds if c is not None] 

                if node.flags == '*m' and not new_path.exists():
                    logger.debug('making uncreated dir', )
                    new_path.mkdir(exist_ok = True, parents = True)

                if pattern_node.type == 'git' and pattern_node.link != '':
                    clone_node(pattern_node.link, new_path)

                if pattern_node.alias != '':
                    logger.debug(f'adding {pattern_node.alias}')

                    logger.debug(f'reversing {node.directory} {pattern_node.directory}')
                    n = ExpressionGenerator.reverse_generator_expression(node.directory, pattern_node.directory)

                    if ( ('*s' in node.flags or '*l' in node.flags) and n == mx_num + 1):
                        self.alias_manager.add( ExpressionGenerator(pattern_node.alias)(1), Path(new_path), is_pathlib = True)
                    elif not ('*s' in node.flags or '*l' in node.flags):
                        self.alias_manager.add( ExpressionGenerator(pattern_node.alias)(n), Path(new_path), is_pathlib = True)
                    
            
            return ret

        else:
            logger.debug('in else', )
            #Before we go to children, let's handle any imports we must
            logger.debug(f'handling imports {node.imports} {node.directory}')

            import_fgraph = None
            if node.imports != '' and Path(node.imports).exists():
                logger.debug(f'importing {node.imports}')
                
                bp, import_fgraph = self.__construct(node.imports, is_output = True)
                import_fgraph.is_root = False

                logger.debug("EEEEEETKJSDLKJSADL", )
            
            if node.flags == '*m' and not new_path.exists():
                new_path.mkdir(exist_ok = True, parents = True)

            if node.type == 'git' and node.link != '':
                clone_node(node.link, new_path)
                
            if node.alias != '':
                logger.debug(f'adding {node.alias}')
                logger.debug(ExpressionGenerator(node.alias)(1))
                self.alias_manager.add(ExpressionGenerator(node.alias)(1), Path(new_path), is_pathlib = True)

            nchild = []
            for child in node.children + [import_fgraph]:
                nchild.extend(self.__expand(child, new_path, step = step))
            
            node.children = [c for c in nchild  if c is not None]

            return [node]

    def __verify(self, node, path = ""):
        '''
            A simple verification method to ensure the basic integrity of a fgraph.

            This method ONLY CHECKS TO SEE IF DIRECTORIES ARE LEGIT. It does nada mas.

            Assumptions:
                - If a node is root, it will be the base directory, and the absolute path will be taken from there.
        '''
        new_path = Path(node.directory) if node.is_root_dir() else Path(path) / node.directory

        logger.debug(f'Testing to see if {new_path} exists')
        if not new_path.exists():
            return [False, None]
        
        new_path = str(new_path)

        if not node.children:
            return [True, node]

        ret = []
        for child in node.children:
            ret.append(self.__verify(child, new_path))


        node.children = [out[1] for out in ret if out[1] != None]

        return [any([out[0] for out in ret]), node]
    

    def __activate(self, node , path = "", mode = "normal"):
        '''
            Activates an fgraph. 
            This means that all variables inside will be collected and factored into the refyre instance.

            The mode parameter defines the behaviour of the method.

            mode = "normal" means normal activation. This is done for input specs, a.k.a. for specs from which we want to 
            COLLECT (retrieve data)

            mode = "deploy" kickstarts deploy activation. Do this for any spec in which we want to TRANSFER / OUTPUT variable
            data somewhere.
        '''

        #Updating the path
        new_path = Path(node.directory) if node.is_root_dir() else Path(path) / node.directory

        if not new_path.exists():
            raise Exception(f"Weird issue caught during activation ... an invalid path appeared - {new_path}")

        new_path = str(new_path)

        if node.name != "":
            logger.info(f'activating {new_path} {node.directory} {node.name}')
            self.__parse_var(node, new_path, mode)

        if node.flags != '':
            #Activate the code manager
            if '*c' in node.flags:
                self.code_manager.add(new_path)


        for child in node.children:
            self.__activate(child, new_path, mode)

    
    def __output(self, node, path = "", mode = "copy"):
        '''
            Outputs an fgraph. 

            By default, this method performs an in-place generation of the file spec. In other words, no directories will be deleted, and it'll try to create 
            any folders that are missing.

            However, if you want the directory structure in the area to look exactly as mentioned in the output spec (no other dirs), you can specify the flag *c 
            using the flags attribute.
        '''
        #Updating the path
        logger.debug(f'in output')
        new_path = Path(node.directory) if node.is_root_dir() else Path(path) / node.directory

        logger.debug(f'mode {node.mode}')
        self.__create_output(node, new_path, mode if node.mode == '' else node.mode)
        new_path = str(new_path)

        for child in node.children:
            self.__output(child, new_path, "copy")
        

    def __clear(self, wipe_proto = True, wipe_vars = True):
        '''
            Wipes the refyre instance clean.
        '''

        if wipe_vars:
            self.variables.clear()
        self.alias_manager.clear()
        self.code_manager.clear()

        self.file_graph = FileGraph()

        if wipe_proto:
            self.file_graph_blueprint = FileGraph()

        gc.collect()

    def __parse_var(self, node, path, mode):
        '''
        Parses all the options in the "names" attribute.

        I expect this method to blow up as this codebase grows (it's an options methods, and you never can have enough options!), 
        so once this goes over 100 lines, we'll move
        it into the fgraph directory.
        '''

        if mode == "normal":
            for v_name in node.name.split(','):
                logger.debug(f'NAME {v_name}')
                name, v = VariableAction(v_name, node, path, self.variables, self.out_temp_var_dict,  True, False, mode = mode)
                logger.debug(f'updated vals {name} {v}')

    def __create_output(self, node, path, mode):
        '''
        Parses all the options in the "names" attribute.

        I expect this method to blow up as this codebase grows (it's an options methods, and you never can have enough options!), 
        so once this goes over 100 lines, we'll move
        it into the fgraph directory.

        mode = "copy" - During output generation, the files are copied over from their original directories 
        mode = "cut"  - During output generation, the files are cut over from their original directories

        '''

        #First of all, let's do the work we need to do 
        logger.debug(f'Creating, {path}')
        path.mkdir(parents = True, exist_ok = True)        
        logger.debug(path.exists())

        if node.name != "":
            for v_name in node.name.split(','):
                logger.debug(f'NAME {v_name}')
                name, v = VariableAction(v_name, node, path, self.variables, self.out_temp_var_dict , False, True, mode = mode)
                logger.debug(f'updated vals {name} {v}')

            
    def __post_generate(self, node, path = "", mode = "copy", flags = ""):
        '''
        Performs any post generation cleanup. For example, it handles the *d flag, which deletes anything
        irrelevant. We do this here to give a chance for any variable data to have been used / moved elsewhere.
        '''

        new_path = Path(node.directory) if node.is_root_dir() else Path(path) / node.directory
        
        logger.debug(f"\n\n {new_path} {node.flags}")

        if node.flags == '*d' or node.flags == "*da" or flags == "*da":

            #We're going to make a list of all the files that should be there, and find all files that aren't in that list 
            logger.debug(f'On {new_path} {node.is_root_dir()}')

            #First, we get a list of all children that aren't roots 
            children = [ (Path(new_path) / pth.directory) for pth in node.children if not pth.is_root_dir()]
            logger.debug(f'non root {children}')

            #Then, we grab all the roots
            children += [ Path(pth.directory) for pth in node.children if pth.is_root_dir()]

            logger.debug(f'all {children}')

            var_pths = []
            if node.name != '':
                #Extract the key information - handling slicing
                name, sliced = VariableParser(node.name, self.variables)

                #If we need to serialize the node, we do it first
                if node.serialize != '':
                    sliced = sliced.rename(ExpressionGenerator(node.serialize))
                    self.variables[name] += sliced

                #Next we grab all the values from our variables that are a part of the dir
                p = sliced.filter(lambda x : new_path in x.parents)


                logger.debug(p)
                var_pths = p.vals()

            logger.debug(f'variable paths {var_pths}')

            #Add 'em up, baby
            tot = children + var_pths
            logger.debug(f'Total {tot}')

            bad_files = []
            for pth in Path(new_path).iterdir():
                if pth not in tot:
                    bad_files.append(pth)
                
            logger.debug(f'Bad files: {bad_files}')
            
            #Snip ... snip ... *ouch* >_<
            for fl in bad_files:
                logger.debug(f'Deleting {fl}')
                if fl.is_dir():
                    shutil.rmtree(str(fl))
                elif fl.is_file():
                    fl.unlink()
                else:
                    logger.error('No clue how to handle this file', )


        new_path = str(new_path)

        for child in node.children:
            self.__post_generate(child, new_path, mode, "*da" if node.flags == "*da" else flags)
        

    def add_spec(self, spec_path, track = False):
        '''
        Adds a spec to parse
        '''

        #Construct an fgraph instance
        proto, graph_to_add = self.__construct(spec_path, is_output = False)
        logger.debug(graph_to_add, )
 
        #Verify the fgraph instance, and then add it in
        verification_successful, verified_graph = self.__verify(graph_to_add)
        if verification_successful:

            logger.debug('Verification successful.', )
            logger.debug(verified_graph, )
            logger.debug('')

            self.file_graph_blueprint.add_graph(proto)
            self.file_graph.add_graph(verified_graph)

            logger.debug('Graph added. Activating variables.\n\n', )
            logger.debug(self.file_graph)
            logger.debug('\n\n')

            self.out_temp_var_dict = self.variables.copy()

            self.__activate(self.file_graph.fgraph_root)

            self.variables = self.out_temp_var_dict

            logger.debug('On standby. Variables: ', )
            logger.debug(self.variables, )
        else:
            logger.error('Vertification failed. Maybe the spec has an invalid dir parameter specified somewhere?', )

        logger.debug('Spec addition complete.\n\n', )
    
    def create_spec(self, spec_path, mode = "cut", track = False):
        '''
        Creates a spec given by the spec path

        spec_path - filepath to the spec path
        mode - 
            - "cut": Any files that are being transferred to create this spec will be cut from their place
            - "copy": Any files that are being transferred to create this spec will be duplicated from their original place
        '''

        assert Path(spec_path).exists(), f"The spec to create at {spec_path} cannot be located"

        logger.debug('Creating spec.', )

        #Construct the spec path
        proto, graph_to_add = self.__construct(spec_path)

        logger.debug(graph_to_add, )
        logger.debug('Graph constructed. Activating variables.', )

        self.out_temp_var_dict = self.variables.copy()

        #Do any actions inside the fgraph
        self.__output(graph_to_add, mode = mode)

        self.variables = self.out_temp_var_dict

        logger.debug('Output complete. Commencing cleanup.', )

        #Do any cleanup inside the fgraph
        self.__post_generate(graph_to_add, mode = mode, flags = graph_to_add.flags)

        if track:
            logger.debug('Tracking requested. Adding to fgraph cluster.', )

            self.file_graph_blueprint.add_graph(proto)
            self.file_graph.add_graph(graph_to_add)

            logger.debug('Graph added.', )
            
    def get_vars(self):
        return self.variables

    def save(self, save_dir = '.'):

        pths = {}

        for var_name in self.variables:
            pths[var_name] = []

            for p in self.variables[var_name]:
                pths[var_name].append(str(p))
            
        save_pth = Path(save_dir) / "refyre_state.json"
        with open(str(save_pth), 'w') as f:
            json.dump(pths, f, indent = 4)
     
    def load(load_filename):

        with open(load_filename, 'r') as f:
            pths = json.load(f)

            variables = {}

            for name in pths:
                variables[name] = FileCluster(values = pths[name], as_pathlib = False)

            return Refyre(variables = variables)
    
    def aliases(self):
        return self.alias_manager

    def step(self):
        '''
        Reconstructs all the variables, generates / functions towards all 
        directories with a '*s' flag in the following ways, based on the flags below:
        
            '*wl' - base write, write latest cluster in the pattern sequence

        Updates to arch - 
        
        Need to maintain a "blueprint" fgraph, pre expansion AND current fgraph

        During step, each directory is created / analyzed
        '''

        logger.debug('\n\n\n STEPPPY', )

        #Prevent conflicts
        self.out_temp_var_dict = self.variables.copy()

        self.__clear(wipe_proto = False)

        self.variables = self.out_temp_var_dict.copy()

        #Reset the state
        self.file_graph_blueprint.fgraph_root, renewed_fgraph = self.__reconstruct(self.file_graph_blueprint.fgraph_root)
        self.variables = self.out_temp_var_dict

        #Add and activate the fgraphs
        self.file_graph.add_graph(renewed_fgraph)


        logger.debug('Step complete', )



from pathlib import Path
import sys
import venv
import site

from refyre.utils import optional_dependencies
from refyre import Refyre
from refyre.config import logger

from pip._internal import main as pip_main
import subprocess
import shutil
import os



def create_virtualenv_and_install_requirements(directory, requirements_file):
    """
    Create a virtual environment in the specified directory using venv.Builder.

    Args:
        directory (str): The directory path where the virtual environment will be created.
    """
    # Create the directory if it doesn't exist
    Path(directory).mkdir(parents=True, exist_ok=True)

    # Set up the builder
    builder = venv.EnvBuilder(with_pip=True)

    # Create the virtual environment
    builder.create(directory)



class RecipePreprocessor:
    def __new__(self, recipe_path):
        '''
            Performs all the recipe path preprocessing, and then 
            returns the Refyre instance
            to manage.

            Returns the refyre instance
        '''
        with optional_dependencies("warn"):
            import yaml 

            assert Path(recipe_path).exists(), f"The input file you were trying to access - {recipe_path} - does NOT exist"

            recipe_dict = None
            with open(recipe_path, 'r') as f:
                recipe_dict = yaml.safe_load(f)
                logger.debug(recipe_dict)
            
            '''
            Currently, a recipe consists of attributes:

                input-spec: a filepath to the input spec
                base-dir: filepath to the base directory / entry point you want your program to see
                env: a description of the environment you want to setup, with two attributes
                    - name: name of the environment
                    - requirements: requirements file

            '''

            logger.debug(recipe_dict)
            logger.debug(recipe_dict['env']['name'])

            #Setup the base directories based on an input spec
            logger.debug(f'Sending to: {recipe_dict["input-spec"]}')


            #Current base directory is assumed to be CWD
            '''
            if 'base-dir' and Path(recipe_dict['base-dir']) != Path.cwd():
                current_base_dir = Path.cwd()

                logger.debug(f'{sys.path}, {current_base_dir.resolve()}')
                sys.path.remove(str(current_base_dir.resolve()))

                sys.path.append(str(Path(recipe_dict['base-dir'].resolve())))
                logger.debug()
            '''

            ref = Refyre.Refyre(input_specs = [recipe_dict["input-spec"]])

            #Create the env if it doesn't exist already or needs to be refreshed
            venv_path = recipe_dict['env']['name']
            if 'env' in recipe_dict and recipe_dict['env']:
                if 'name' in recipe_dict['env']:
                    if not Path(recipe_dict['env']['name']).exists():
                        create_virtualenv_and_install_requirements(venv_path, recipe_dict['env']['requirements'])
                    elif 'refresh' in recipe_dict['env'] and recipe_dict['env']['refresh']:
                        shutil.rmtree(recipe_dict['env']['name'])
                        create_virtualenv_and_install_requirements(recipe_dict['env']['name'], recipe_dict['env']['requirements'])


            return ref
            




import re 
from refyre.config import logger


class ExpressionGenerator:
    '''
        The ExpressionGenerator class takes in an 
        input generator string, such as test$.txt, and returns a
        lambda function that can be used to evaluate the expressions at various points.

        For example, test$.txt returns a lambda function, that when called at i = 1, returns test1.txt

        This class is a static class; it's sole purpose is to create an organized location
        to generate complex expressions.

        For now, only numbers are supported. You can generate number expressions using the '$' symbol 
        as shown above.

    '''
    def generate_expression(string):
        # Replace '$' with '{}' to use it as a placeholder for string formatting
        lambda_function = lambda i: string.replace('$', str(i))

        return lambda_function
    
    def is_valid_genexp(expression):
        pattern = r'^[a-zA-Z0-9/\-_$]+$'
        logger.debug(f'Pattern: {pattern}')
        logger.debug(f'Ret: {bool(re.match(pattern, expression))}')
        return bool(re.match(pattern, expression))
    
    def is_null_generator(expression):
        '''
            By "null" generator, I just mean a genexp that 
            doesn't have any special characters, i.e. one that will stay the same
        '''
        logger.debug(f"Checking: , {expression}, {'$' in expression}")
        return not ('$' in expression)
    
    def convert_generator_expression(expression):
        escaped_expression = re.escape(expression)
        regex_pattern = escaped_expression.replace(r'\$', r'(\d+)')
        return regex_pattern
    
    def reverse_generator_expression(expression, string):
        regex_pattern = ExpressionGenerator.convert_generator_expression(expression)

        logger.debug(f'{expression}, {regex_pattern}')
        match = re.match(regex_pattern, string)
        if match:
            logger.debug(f"{match.group(1)}, exp,  {expression}, str, {string}")
            return int(match.group(1))
        else:
            return None
        
    def __new__(self, string):
        '''
            Hijack the '__new__' method (usually used in constructors) 
            and make it do our bidding >:)

            Takes in an expression, return a lambda function.
        '''

        return ExpressionGenerator.generate_expression(string)


import sys 
from pathlib import Path

class CodeManager:

    def __init__(self):
        self.module_dirs = []
        self.base_sys_paths = list(sys.path)

    def add(self, filepath):
        p = Path(filepath)
        self.module_dirs.append(str(p.parent)) 
        sys.path.append(str(p.parent))

    def clear(self):
        self.module_dirs.clear()


    



from pathlib import Path  
from refyre.config import logger

def insert(fdct, k, v):
    if k not in fdct:
        fdct[k] = set()
    
    fdct[k].add(v)

class Broadcaster:

    files_dict = {}
    ignore_list = []

    def __init__(self, *a, **kw):
        self.conf_args = a 
        self.conf_kw   = kw
    
    @classmethod
    def clear(cls):
        cls.files_dict.clear()

    @classmethod
    def add(cls, id, cluster_vals):
        for v in cluster_vals:
            insert(cls.files_dict, str(v), id) 

        
    @classmethod
    def release(cls, cluster_id):
        logger.debug(f'released {cluster_id}')
        for k in cls.files_dict:
            logger.debug(f'{k}, {cluster_id}, {cls.files_dict[k]}')
            cls.files_dict[k].discard(cluster_id)

        if cluster_id in cls.ignore_list:
            cls.ignore_list.remove(cluster_id)

    @classmethod
    def unlink(cls, cluster_id):
        if cluster_id not in cls.ignore_list:
            cls.ignore_list.append(cluster_id)
        

    def __call__(self, func):
        
        def wrapper(instance, *args, **kwargs):
            '''
                change_arr: List[Tuple(Str, Str)]
                    - Lists the Str that the path started out as, and what it became. 
                    - If second tuple None, it implies a path was deleted

            '''

            if instance.id not in self.ignore_list:
                instance.values = [v for v in list(dict.fromkeys(instance.values)) if v.exists() ]

            change_arr, out = func(instance, *args, **kwargs)

            if instance.id in self.ignore_list:
                return out


            logger.debug(f"CHANGE {change_arr}")
            logger.debug(out)

            clusters = instance.all_clusters()

            logger.debug(self.files_dict)
            self.files_dict = {k:self.files_dict[k] for k in self.files_dict if len(self.files_dict[k]) > 0}

            for before, after in change_arr:

                assert before != None 

                if before not in self.files_dict:
                    continue

                #Extract the dictionary values for before
                before_vals = self.files_dict.pop(before)

                #Assert no instance of after exists (otherwise a duplicate) (either None, or not in files_dict)
                logger.debug(f'{before}, {after}, {self.files_dict}')

                #Move everything to after & make the mods
                if after:
                    logger.debug('after not none')
                    if after in self.files_dict:
                        logger.debug(f'heeres {self.files_dict[after]}')
                    self.files_dict[after] = before_vals.union(self.files_dict[after]) if after in self.files_dict else before_vals
                    for cluster_id in before_vals:
                        cluster = clusters[cluster_id]()
                        if cluster_id != instance.id and Path(before) in cluster.values:
                            cluster.values[cluster.values.index(Path(before))] = Path(after)
                        
                else:
                    for cluster_id in before_vals:
                        if cluster_id != instance.id:
                            cluster = clusters[cluster_id]()
                            cluster.values = [v for v in cluster.values if str(v) != before] 

            self.files_dict = {k:self.files_dict[k] for k in self.files_dict if len(self.files_dict[k]) > 0 and Path(k).exists()}
            Broadcaster.files_dict = self.files_dict

            instance.values = [v for v in list(dict.fromkeys(instance.values)) if v.exists() ]
            logger.debug(f"RETURNED {instance.values} {self.files_dict} {Broadcaster.files_dict}")

            return out
            
        return wrapper 



from pathlib import Path
from itertools import product
from refyre.config import logger

from .Cluster import FileCluster
from .ClusterIterator import FileClusterIterator

#Store a weak reference to the object
import weakref

class AssociationCluster:

    #Counter to manage IDs
    GLOBAL_COUNTER = 0

    #Stores references to each AssociationCluster; in the future, this can be used for tracking & mem management
    clusters = {}

    def __init__(self, starting_vals = None, mapper_function = lambda x : x, one_to_one = False , as_pathlib = False, input_vars = []):
        self.values = []


        if starting_vals:
            out = []
            if not as_pathlib:
                for p in starting_vals:
                    t = (Path(v) for v in p)
                    out += t
            
            self.values += out

        #ID the variable
        self.id = AssociationCluster.GLOBAL_COUNTER
        AssociationCluster.GLOBAL_COUNTER += 1

        #Track the cluster without affecting garbage collection
        self.weak_ref = weakref.ref(self)
        AssociationCluster.clusters[self.id] = self.weak_ref

        if input_vars:
            input_vars = [v.vals() for v in input_vars]
            iter_form = product(*input_vars) if not one_to_one else zip(*input_vars)
            for permutation in iter_form:
                if mapper_function(permutation):
                    self.values.append(permutation)
        
    
    def clusterify(self, mapper_function = None):
        single_values = []

        for val_set in self.values:
            if mapper_function:
                m = mapper_func(*val_set)
                if isinstance(m, list):
                    single_values.extend(m)
                else:
                    single_values.append(m)
            else:
                for v in val_set:
                    single_values.append(v) 
        
        return FileCluster(values = single_values, as_pathlib = True)
        
    def vals(self):
        return self.values

    def __len__(self):
        return len(self.values) 
    
    def __contains__(self, k):
        return k in self.values

    def __iter__(self):
        return FileClusterIterator(self)

    def __repr__(self):
        return f"AssociationCluster(values = {self.values})"

    def __copy__(self):
        return AssociationCluster(values = self.values, as_pathlib = True)
    
    def __getitem__(self, key):
        if isinstance(key, slice):
            return AssociationCluster(values = self.values[key.start:key.stop:key.step], as_pathlib = True)
        elif isinstance(key, int):
            return AssociationCluster(values = [self.values[key]], as_pathlib = True)
    
    def __del__(self):
        logger.debug(f'deleting {self.id}')
        if self.id in AssociationCluster.clusters:
            AssociationCluster.clusters.pop(self.id)
    
    @classmethod
    def cleanup(cls):
        for k in list(AssociationCluster.clusters):
            obj = AssociationCluster.clusters[k]
            if obj() is None:
                AssociationCluster.clusters.pop(k)

    def all_clusters(self):
        return AssociationCluster.clusters

    def __eq__(self, other):
        if other is None or type(other) != AssociationCluster:
            return False
        return set(self.values) == set(other.values)
    
    def __add__(self, other):
        '''
            other - another AssociationCluster

            Returns a file cluster with all the combined values
        '''
        #Ensure the input_data is as desired
        assert type(other) == AssociationCluster

        new_values = self.values + other.values

        return AssociationCluster(values = new_values, as_pathlib = True)


import re
import fnmatch

'''
Utility methods for regex related things 
'''

def is_valid_regex(string):
    try:
        re.compile(fnmatch.translate(string))
        return True
    except re.error:
        return False

def get_optimal_pattern(string):
    try:
        log('string is a regex', string )
        re.compile(string)
        return string
    except re.error:
        return False
    
    try:
        log('string is a glob', string )
        re.compile(fnmatch.translate(string))
        return fnmatch.translate(string)
    except re.error:    
        return None

def extract_numbers(string):
    pattern = r'(\d+)(?:-(\d+))?$'
    match = re.match(pattern, string)
    if match:

        if not match.group(2):
            return 0, int(match.group(1))

        start = int(match.group(1))
        end = int(match.group(2))

        return start, end
    else:
        return None


from pathlib import Path
from refyre.config import logger
import json
import shutil


class AutoCache:

    cache_files = set()

    def __init__(self, cache_dir = '.refyre_cache'):
        self.cache_dir = Path(cache_dir)
        
        self.cache_dir.mkdir(parents = True, exist_ok = True)

        if (self.cache_dir / '.refyre_cache.json').exists():
            self.cache_files = set(self.load(self.cache_dir / '.refyre_cache.json'))
        else:
            self.cache_files = set()


    def cache(self, var):

        #Delete all the files in the cache with the same names
        for fl in var:
            if fl.name in self.cache_files:
                if fl.is_dir():
                    shutil.rmtree(str(self.cache_dir / fl.name))
                elif fl.is_file():
                    (self.cache_dir / fl.name).unlink()

        logger.debug(f'CACHED {self.cache_files}')
        for fl in var:
            if fl.name not in self.cache_files:
                self.cache_files.add(fl.name)

        logger.debug(f'CACHED {self.cache_files}')
        
        return var.copy(str(self.cache_dir))
    
    def decache(self, var, dest_dir = '.'):
        #Delete all the files in the cache with the same names
        nvals = []
        for fl in var:
            if fl.name in self.cache_files:
                nvals.append(self.cache_dir / fl.name)
                self.cache_files.discard(fl.name)

        from .Cluster import FileCluster
        return FileCluster(values = nvals, as_pathlib = True).move(dest_dir)

    def load(self, fp):
        with open(str(fp), 'r') as f:
            dct = json.load(f)
        return dct['files']

    def save(self, pth = None):
        if self.cache_dir.exists():
            with open(str(self.cache_dir / '.refyre_cache.json') if not pth else pth, 'w') as f:
                json.dump({'files' : list(self.cache_files) }, f)

    def contains(self, fp):
        return Path(fp).name in self.cache_files
        
    def __del__(self):
        self.save()
    
    @classmethod
    def obliterate(cls, cache_instance):
        """
        Automatically wipe the entire cache object
        """
        shutil.rmtree(str(cache_instance.cache_dir))
    
    def __len__(self):
        return len(self.cache_files) - 1




import argparse 
import sys
from refyre import Refyre
from refyre.core import RecipePreprocessor

#Instantiate an argument parser
def main(argv = None):
    parser = argparse.ArgumentParser(
        prog = "cli",
        description = "refyre v.0.0.1 CLI",
    )

    #Add positional arguments for input specs
    parser.add_argument("-i", '--input', help = "filepaths to the input specs, seperated by spaces", nargs='+')
    parser.add_argument("-o", '--output', help = "filepaths to the output specs, seperated by spaces", nargs='+')
    parser.add_argument("-r", '--recipe', help = "path to recipe / workflow file you want refyre to execute")

    #Add positional arguments for saving refyre state
    parser.add_argument("-s", '--save', help = "save the refyre state locally")

    args = parser.parse_args(argv)

    if args.recipe != None:
        RecipePreprocessor(args.recipe)

    ref = Refyre(input_specs = args.input, output_specs = args.output)

    if args.save:
        ref.save()

if __name__ == '__main__':
    sys.exit(main(sys.argv[1:]))


from .VariableParser import VariableParser
from refyre.cluster import FileCluster
from refyre.config import logger

import re

class VariableAction:
    '''
    External class that handles all of the variable updates,
    and provides an interface for the rest of the code to read & write 
    to variables

    Together, with the VariableParser, this class assembles "intent" that can 
    be used in specs.
    '''

    def __new__(self, var_name , node, path, variable_dict, out_variable_dict , is_read, is_write, mode = ""):
        '''
        Reads in all parameters, and returns the name of the updated 
        variable

        Ironically, the is_read and is_write are defined in opposite.

        is_read refers to whether we are trying to read a value into the variable (which really just means "write"), but
        I'm keeping the naming this way since we're mostly doing the writing in specs focused on reading

        Similarly, is_write is for output specs
        '''

        logger.debug(mode)
        #Currently, this operator pattern supports only + and "" (appending and defining)
        options_pattern = r'^(\+)?(.*)$' 

        match = re.match(options_pattern, var_name)
        operator = match.group(1) or ""
        name = match.group(2)

        #If the operator is "", a read command is sought
        '''
        Each of these commands have dual meanings - 

        "" in read means to read all the data to the variable
        "" in write means to 


        '''

        if is_read:
            if operator == "":
                out_variable_dict[name] = FileCluster(input_paths = [path], input_patterns = [node.pattern], recursive = "*r" in node.flags)

                if '*f' in node.flags:
                    out_variable_dict[name] = out_variable_dict[name].filter(lambda x : x.is_file())

                #*f flag - only keep directories
                if '*d' in node.flags:
                    out_variable_dict[name] = out_variable_dict[name].filter(lambda x : x.is_dir())

            elif operator == "+":
                
                p = FileCluster(input_paths = [path], input_patterns = [node.pattern], recursive = "*r" in node.flags)

                if '*f' in node.flags:
                    p = p.filter(lambda x : x.is_file())

                #*f flag - only keep directories
                if '*d' in node.flags:
                    p = p.filter(lambda x : x.is_dir())

                #Add a conditional in case they never defined a variable before
                out_variable_dict[name] = (p if name not in out_variable_dict else out_variable_dict[name] + p)


            return name, out_variable_dict[name] 
        
        if is_write:

            logger.debug(f'{name} {variable_dict}')
            name, sliced, start, stop, step = VariableParser(name, variable_dict)

            #The refresher method will eliminate the invalid paths, so let's add back in the new paths at their original positions
            v = list(variable_dict[name].vals())

            if mode == "copy":
                sliced = sliced.copy(path)
            
            elif mode == "cut":
                sliced = sliced.move(path)
            


            if operator == '':
                s_v = sliced.vals()
                for ind_s, ind_v in enumerate(range(start, stop, step)):
                    v[ind_v] = s_v[ind_s]
                logger.debug(f'updated vals {v}')

                out_variable_dict[name] = variable_dict[name]
            elif operator == "+":
                logger.debug(f"APPENDING {variable_dict[name]} {v}")


                out_variable_dict[name] += sliced

            return name, out_variable_dict[name]

        


from refyre.utils import optional_dependencies 

with optional_dependencies("warn"):
    import torch
    from torch.utils.data import Dataset

    class TorchStack(Dataset):
        def __init__(self, association):
            self.association = association

        def __len__(self):
            return len(self.association)

        def items(self, k):
            '''
                If you are inheriting this dataset to make your own, 
                use this method to retrieve the data for the __getitem__ method
            '''
            return self.association[k]

        def __getitem__(self, idx):
            return self.items(idx)


import logging 
from logging.config import fileConfig
from pathlib import Path

'''
Setup for the refyre logger
'''

logging.basicConfig(filename="refyre.log",
                    format='%(levelname)s %(message)s',
                    filemode='w')

logger = logging.getLogger()
 
# Setting the threshold of logger to DEBUG
logger.setLevel(logging.CRITICAL)


import fnmatch
import re
from .ExpressionGenerator import ExpressionGenerator
from refyre.config import logger

class PatternGenerator:
    '''
        The pattern generator is designed to parse all pattern based 
        strings and return regex versions of them that can be used to 
        pattern match.

        We created the generator to help differentiate between glob patterns and
        regex. Most of the time, people use glob, but sometimes the power of regex 
        matching does come in handy. 

        A glob string will come in the normal form. A regex string will start with an r before the actual pattern.

        ex: "a?.txt" is a glob pattern. "ra?.txt" is regex.

    '''
    REGEX_STARTER = 'r!'
    GLOB_STARTER = 'g!'

    def is_valid_regex(pattern_string):
        logger.debug(pattern_string)
        if pattern_string.startswith(PatternGenerator.REGEX_STARTER):
            try:
                re.compile(pattern_string[2:])
            except re.error as e:
                logger.error(e)
                return False
        elif pattern_string.startswith(PatternGenerator.GLOB_STARTER):
            try:
                re.compile(fnmatch.translate(pattern_string[2:]))
            except re.error as e:
                logger.error(e)
                return False
        else:
            if not PatternGenerator.is_valid_genexp(pattern_string):
                logger.debug('not a valid')
                return False
        return True

    def is_valid_genexp(expression):
        return ExpressionGenerator.is_valid_genexp(expression)
    
    def convert_generator_expression(expression):
        return ExpressionGenerator.convert_generator_expression(expression)

    def get_pattern_type(pattern_string):
        if pattern_string.startswith(PatternGenerator.REGEX_STARTER):
            return "regex"

        if pattern_string.startswith(PatternGenerator.GLOB_STARTER):
            return "glob"
        
        #Handle generator expressions
        if not ExpressionGenerator.is_null_generator(pattern_string):
            logger.debug('here, we think it is a expgen')
            return "generator_expression"

        #Else, the string is a normal string
        return "normal_string" 

    def __new__(self, pattern_string, given_type = None):
        '''
            Takes in an input pattern string and returns the appropriate regex pattern.

        '''
        if given_type == 'regex' or given_type == 'r':
            return pattern_string

        if given_type == 'glob' or given_type == 'g':
            return fnmatch.translate(pattern_string)

        if given_type == 'expression_generator' or given_type == 'e':
            return ExpressionGenerator.convert_generator_expression(pattern_string)
        
        if given_type == 'normal' or given_type == 'n':
            return None 

        #If the type isn't given, we'll have to infer it
        if pattern_string.startswith(PatternGenerator.REGEX_STARTER):
            logger.debug('regex')
            return pattern_string[2:]

        if pattern_string.startswith(PatternGenerator.GLOB_STARTER):
            logger.debug('glob')
            return fnmatch.translate(pattern_string[2:])
        
        #Handle generator expressions
        if not ExpressionGenerator.is_null_generator(pattern_string):
            return ExpressionGenerator.convert_generator_expression(pattern_string)

        #Else, the string is a normal string
        return pattern_string